"""
Module detecting unimplemented interfaces

Collect all the interfaces
Check for contracts which implement all interface functions but do not explicitly derive from those interfaces.
"""
from typing import List
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.core.declarations.contract import Contract
from slither.utils.output import Output


class MissingInheritance(AbstractDetector):
    """
    Unimplemented interface detector
    """

    ARGUMENT = "missing-inheritance"
    HELP = "Missing inheritance"
    IMPACT = DetectorClassification.INFORMATIONAL
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#missing-inheritance"
    WIKI_TITLE = "Missing inheritance"
    WIKI_DESCRIPTION = "Detect missing inheritance."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
interface ISomething {
    function f1() external returns(uint);
}

contract Something {
    function f1() external returns(uint){
        return 42;
    }
}
```
`Something` should inherit from `ISomething`. 
"""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Inherit from the missing interface or contract."

    @staticmethod
    def detect_unimplemented_interface(
        contract: Contract, interfaces: List[Contract]
    ) -> List[Contract]:
        """
        Detects if contract intends to implement one of the interfaces but does not explicitly do so by deriving from it
        :param contract: The contract to check
        :param interfaces: List of all the interfaces
        :return: Interfaces likely intended to implement by the contract
        """

        intended_interfaces: List[Contract] = []
        sigs_contract = {f.full_name for f in contract.functions_entry_points}

        if not sigs_contract:
            return intended_interfaces

        for interface in interfaces:
            # If contract already inherits from interface, skip that interface
            if interface in contract.inheritance:
                continue

            sigs_interface = {f.full_name for f in interface.functions_entry_points}

            # Contract should implement all the functions of the intended interface
            if not sigs_interface.issubset(sigs_contract):
                continue

            # A parent contract inherited should not implement a superset of intended interface
            # This is because in the following:
            # interface ERC20_interface:
            #    - balanceOf(uint) -> uint
            #    - transfer(address, address) -> bool
            # contract ForeignToken:
            #    - balanceOf(uint) -> uint
            # contract MyERC20Implementation is ERC20_interface
            #
            # We do not want MyERC20Implementation to be declared as missing ForeignToken interface
            intended_interface_is_subset_parent = False
            for parent in contract.inheritance:
                sigs_parent = {f.full_name for f in parent.functions_entry_points}
                if sigs_interface.issubset(sigs_parent):
                    intended_interface_is_subset_parent = True
                    break

            if not intended_interface_is_subset_parent:
                # Should not be a subset of an earlier determined intended_interface or derive from it
                intended_interface_is_subset_intended = False
                for intended_interface in list(intended_interfaces):
                    sigs_intended_interface = {
                        f.full_name for f in intended_interface.functions_entry_points
                    }
                    if (
                        sigs_interface.issubset(sigs_intended_interface)
                        or interface in intended_interface.inheritance
                    ):
                        intended_interface_is_subset_intended = True
                        break

                    # If superset of an earlier determined intended_interface or derives from it,
                    # remove the intended_interface
                    if (
                        sigs_intended_interface.issubset(sigs_interface)
                        or intended_interface in interface.inheritance
                    ):
                        intended_interfaces.remove(intended_interface)

                if not intended_interface_is_subset_intended:
                    intended_interfaces.append(interface)

        return intended_interfaces

    def _detect(self) -> List[Output]:
        """Detect unimplemented interfaces
        Returns:
            list: {'contract'}
        """

        # Collect all the interfaces
        # Here interface can be "interface" from solidity, or contracts with only functions declaration
        # Skip interfaces without functions
        interfaces = [
            contract
            for contract in self.compilation_unit.contracts
            if contract.is_signature_only()
            and any(not f.is_constructor_variables for f in contract.functions)
        ]

        # Check derived contracts for missing interface implementations
        results = []
        for contract in self.compilation_unit.contracts_derived:
            # Skip interfaces
            if contract in interfaces:
                continue
            intended_interfaces = self.detect_unimplemented_interface(contract, interfaces)
            for interface in intended_interfaces:
                info: DETECTOR_INFO = [contract, " should inherit from ", interface, "\n"]
                res = self.generate_result(info)
                results.append(res)
        return results



from pathlib import Path
from typing import List, Optional, Tuple
import shutil
import sys
import sysconfig

from slither.utils.colors import yellow, green, red


def path_is_relative_to(path: Path, relative_to: Path) -> bool:
    """
    Check if a path is relative to another one.

    Compatibility wrapper for Path.is_relative_to
    """
    if sys.version_info >= (3, 9, 0):
        return path.is_relative_to(relative_to)

    path_parts = path.resolve().parts
    relative_to_parts = relative_to.resolve().parts

    if len(path_parts) < len(relative_to_parts):
        return False

    for (a, b) in zip(path_parts, relative_to_parts):
        if a != b:
            return False

    return True


def check_path_config(name: str) -> Tuple[bool, Optional[Path], List[Path]]:
    """
    Check if a given Python binary/script is in PATH.
    :return: Returns if the binary on PATH corresponds to this installation,
             its path (if present), and a list of possible paths where this
             binary might be found.
    """
    binary_path = shutil.which(name)
    possible_paths = []

    for scheme in sysconfig.get_scheme_names():
        script_path = Path(sysconfig.get_path("scripts", scheme))
        purelib_path = Path(sysconfig.get_path("purelib", scheme))
        script_binary_path = shutil.which(name, path=script_path)
        if script_binary_path is not None:
            possible_paths.append((script_path, purelib_path))

    binary_here = False
    if binary_path is not None:
        binary_path = Path(binary_path)
        this_code = Path(__file__)
        this_binary = list(filter(lambda x: path_is_relative_to(this_code, x[1]), possible_paths))
        binary_here = len(this_binary) > 0 and all(
            path_is_relative_to(binary_path, script) for script, _ in this_binary
        )

    return binary_here, binary_path, list(set(script for script, _ in possible_paths))


def check_slither_path(**_kwargs) -> None:
    binary_here, binary_path, possible_paths = check_path_config("slither")
    show_paths = False

    if binary_path:
        print(green(f"`slither` found in PATH at `{binary_path}`."))
        if binary_here:
            print(green("Its location matches this slither-doctor installation."))
        else:
            print(
                yellow(
                    "This path does not correspond to this slither-doctor installation.\n"
                    + "Double-check the order of directories in PATH if you have several Slither installations."
                )
            )
            show_paths = True
    else:
        print(red("`slither` was not found in PATH."))
        show_paths = True

    if show_paths:
        print()
        print("Consider adding one of the following directories to PATH:")
        for path in possible_paths:
            print(f"  * {path}")



from fractions import Fraction
from typing import Union
from Crypto.Hash import keccak

from slither.core import expressions
from slither.core.expressions import (
    BinaryOperationType,
    Literal,
    UnaryOperationType,
    Identifier,
    BinaryOperation,
    UnaryOperation,
    TupleExpression,
    TypeConversion,
    CallExpression,
)
from slither.core.variables import Variable
from slither.utils.integer_conversion import convert_string_to_fraction, convert_string_to_int
from slither.visitors.expression.expression import ExpressionVisitor
from slither.core.solidity_types.elementary_type import ElementaryType


class NotConstant(Exception):
    pass


KEY = "ConstantFolding"

CONSTANT_TYPES_OPERATIONS = Union[
    Literal, BinaryOperation, UnaryOperation, Identifier, TupleExpression, TypeConversion
]


def get_val(expression: CONSTANT_TYPES_OPERATIONS) -> Union[bool, int, Fraction, str]:
    val = expression.context[KEY]
    # we delete the item to reduce memory use
    del expression.context[KEY]
    return val


def set_val(expression: CONSTANT_TYPES_OPERATIONS, val: Union[bool, int, Fraction, str]) -> None:
    expression.context[KEY] = val


class ConstantFolding(ExpressionVisitor):
    def __init__(
        self, expression: CONSTANT_TYPES_OPERATIONS, custom_type: Union[str, "ElementaryType"]
    ) -> None:
        if isinstance(custom_type, str):
            custom_type = ElementaryType(custom_type)
        self._type: ElementaryType = custom_type
        super().__init__(expression)

    @property
    def expression(self) -> CONSTANT_TYPES_OPERATIONS:
        # We make the assumption that the expression is always a CONSTANT_TYPES_OPERATIONS
        # Other expression are not supported for constant unfolding
        return self._expression  # type: ignore

    def result(self) -> "Literal":
        value = get_val(self.expression)
        if isinstance(value, Fraction):
            value = int(value)
            # emulate 256-bit wrapping
            if str(self._type).startswith("uint"):
                value = value & (2**256 - 1)
        return Literal(value, self._type)

    # pylint: disable=import-outside-toplevel
    def _post_identifier(self, expression: Identifier) -> None:
        from slither.core.declarations.solidity_variables import SolidityFunction

        if isinstance(expression.value, Variable):
            if expression.value.is_constant:
                expr = expression.value.expression
                # assumption that we won't have infinite loop
                # Everything outside of literal
                if isinstance(
                    expr,
                    (BinaryOperation, UnaryOperation, Identifier, TupleExpression, TypeConversion),
                ):
                    cf = ConstantFolding(expr, self._type)
                    expr = cf.result()
                assert isinstance(expr, Literal)
                set_val(expression, convert_string_to_int(expr.converted_value))
            else:
                raise NotConstant
        elif isinstance(expression.value, SolidityFunction):
            set_val(expression, expression.value)
        else:
            raise NotConstant

    # pylint: disable=too-many-branches,too-many-statements
    def _post_binary_operation(self, expression: BinaryOperation) -> None:
        expression_left = expression.expression_left
        expression_right = expression.expression_right
        if not isinstance(
            expression_left,
            (Literal, BinaryOperation, UnaryOperation, Identifier, TupleExpression, TypeConversion),
        ):
            raise NotConstant
        if not isinstance(
            expression_right,
            (Literal, BinaryOperation, UnaryOperation, Identifier, TupleExpression, TypeConversion),
        ):
            raise NotConstant
        left = get_val(expression_left)
        right = get_val(expression_right)

        if (
            expression.type == BinaryOperationType.POWER
            and isinstance(left, (int, Fraction))
            and isinstance(right, (int, Fraction))
        ):
            set_val(expression, left**right)  # type: ignore
        elif (
            expression.type == BinaryOperationType.MULTIPLICATION
            and isinstance(left, (int, Fraction))
            and isinstance(right, (int, Fraction))
        ):
            set_val(expression, left * right)
        elif (
            expression.type == BinaryOperationType.DIVISION
            and isinstance(left, (int, Fraction))
            and isinstance(right, (int, Fraction))
        ):
            # TODO: maybe check for right + left to be int to use // ?
            set_val(expression, left // right if isinstance(right, int) else left / right)
        elif (
            expression.type == BinaryOperationType.MODULO
            and isinstance(left, (int, Fraction))
            and isinstance(right, (int, Fraction))
        ):
            set_val(expression, left % right)
        elif (
            expression.type == BinaryOperationType.ADDITION
            and isinstance(left, (int, Fraction))
            and isinstance(right, (int, Fraction))
        ):
            set_val(expression, left + right)
        elif (
            expression.type == BinaryOperationType.SUBTRACTION
            and isinstance(left, (int, Fraction))
            and isinstance(right, (int, Fraction))
        ):
            set_val(expression, left - right)
        # Convert to int for operations not supported by Fraction
        elif expression.type == BinaryOperationType.LEFT_SHIFT:
            set_val(expression, int(left) << int(right))
        elif expression.type == BinaryOperationType.RIGHT_SHIFT:
            set_val(expression, int(left) >> int(right))
        elif expression.type == BinaryOperationType.AND:
            set_val(expression, int(left) & int(right))
        elif expression.type == BinaryOperationType.CARET:
            set_val(expression, int(left) ^ int(right))
        elif expression.type == BinaryOperationType.OR:
            set_val(expression, int(left) | int(right))
        elif expression.type == BinaryOperationType.LESS:
            set_val(expression, int(left) < int(right))
        elif expression.type == BinaryOperationType.LESS_EQUAL:
            set_val(expression, int(left) <= int(right))
        elif expression.type == BinaryOperationType.GREATER:
            set_val(expression, int(left) > int(right))
        elif expression.type == BinaryOperationType.GREATER_EQUAL:
            set_val(expression, int(left) >= int(right))
        elif expression.type == BinaryOperationType.EQUAL:
            set_val(expression, int(left) == int(right))
        elif expression.type == BinaryOperationType.NOT_EQUAL:
            set_val(expression, int(left) != int(right))
        # Convert boolean literals from string to bool
        elif expression.type == BinaryOperationType.ANDAND:
            set_val(expression, left == "true" and right == "true")
        elif expression.type == BinaryOperationType.OROR:
            set_val(expression, left == "true" or right == "true")
        else:
            raise NotConstant

    def _post_unary_operation(self, expression: UnaryOperation) -> None:
        # Case of uint a = -7; uint[-a] arr;
        if expression.type == UnaryOperationType.MINUS_PRE:
            expr = expression.expression
            # Everything outside of literal
            if isinstance(
                expr, (BinaryOperation, UnaryOperation, Identifier, TupleExpression, TypeConversion)
            ):
                cf = ConstantFolding(expr, self._type)
                expr = cf.result()
            assert isinstance(expr, Literal)
            set_val(expression, -convert_string_to_fraction(expr.converted_value))
        else:
            raise NotConstant

    def _post_literal(self, expression: Literal) -> None:
        if str(expression.type) == "bool":
            set_val(expression, expression.converted_value)
        elif str(expression.type) == "string":
            set_val(expression, expression.converted_value)
        else:
            try:
                set_val(expression, convert_string_to_fraction(expression.converted_value))
            except ValueError as e:
                raise NotConstant from e

    def _post_assignement_operation(self, expression: expressions.AssignmentOperation) -> None:
        raise NotConstant

    def _post_call_expression(self, expression: expressions.CallExpression) -> None:
        called = get_val(expression.called)
        args = [get_val(arg) for arg in expression.arguments]
        if called.name == "keccak256(bytes)":
            digest = keccak.new(digest_bits=256)
            digest.update(str(args[0]).encode("utf-8"))
            set_val(expression, digest.digest())
        else:
            raise NotConstant

    def _post_conditional_expression(self, expression: expressions.ConditionalExpression) -> None:
        raise NotConstant

    def _post_elementary_type_name_expression(
        self, expression: expressions.ElementaryTypeNameExpression
    ) -> None:
        raise NotConstant

    def _post_index_access(self, expression: expressions.IndexAccess) -> None:
        raise NotConstant

    def _post_member_access(self, expression: expressions.MemberAccess) -> None:
        raise NotConstant

    def _post_new_array(self, expression: expressions.NewArray) -> None:
        raise NotConstant

    def _post_new_contract(self, expression: expressions.NewContract) -> None:
        raise NotConstant

    def _post_new_elementary_type(self, expression: expressions.NewElementaryType) -> None:
        raise NotConstant

    def _post_tuple_expression(self, expression: expressions.TupleExpression) -> None:
        if expression.expressions:
            if len(expression.expressions) == 1:
                first_expr = expression.expressions[0]
                if not isinstance(
                    first_expr,
                    (
                        Literal,
                        BinaryOperation,
                        UnaryOperation,
                        Identifier,
                        TupleExpression,
                        TypeConversion,
                    ),
                ):
                    raise NotConstant
                cf = ConstantFolding(first_expr, self._type)
                expr = cf.result()
                assert isinstance(expr, Literal)
                set_val(expression, convert_string_to_fraction(expr.converted_value))
                return
        raise NotConstant

    def _post_type_conversion(self, expression: expressions.TypeConversion) -> None:
        expr = expression.expression
        if not isinstance(
            expr,
            (
                Literal,
                BinaryOperation,
                UnaryOperation,
                Identifier,
                TupleExpression,
                TypeConversion,
                CallExpression,
            ),
        ):
            raise NotConstant
        cf = ConstantFolding(expr, self._type)
        expr = cf.result()
        assert isinstance(expr, Literal)
        if str(expression.type).startswith("uint") and isinstance(expr.value, bytes):
            value = int.from_bytes(expr.value, "big")
        elif str(expression.type).startswith("byte") and isinstance(expr.value, int):
            value = int.to_bytes(expr.value, 32, "big")
        else:
            value = convert_string_to_fraction(expr.converted_value)
        set_val(expression, value)



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="slither-analyzer",
    description="Slither is a Solidity and Vyper static analysis framework written in Python 3.",
    url="https://github.com/crytic/slither",
    author="Trail of Bits",
    version="0.10.1",
    packages=find_packages(),
    python_requires=">=3.8",
    install_requires=[
        "packaging",
        "prettytable>=3.3.0",
        "pycryptodome>=3.4.6",
        "crytic-compile>=0.3.6,<0.4.0",
        # "crytic-compile@git+https://github.com/crytic/crytic-compile.git@master#egg=crytic-compile",
        "web3>=6.0.0",
        "eth-abi>=4.0.0",
        "eth-typing>=3.0.0",
        "eth-utils>=2.1.0",
    ],
    extras_require={
        "lint": [
            "black==22.3.0",
            "pylint==2.13.4",
        ],
        "test": [
            "pytest",
            "pytest-cov",
            "pytest-xdist",
            "deepdiff",
            "numpy",
            "coverage[toml]",
            "filelock",
            "pytest-insta",
        ],
        "doc": [
            "pdoc",
        ],
        "dev": [
            "slither-analyzer[lint,test,doc]",
            "openai",
        ],
    },
    license="AGPL-3.0",
    long_description=long_description,
    long_description_content_type="text/markdown",
    entry_points={
        "console_scripts": [
            "slither = slither.__main__:main",
            "slither-check-upgradeability = slither.tools.upgradeability.__main__:main",
            "slither-find-paths = slither.tools.possible_paths.__main__:main",
            "slither-simil = slither.tools.similarity.__main__:main",
            "slither-flat = slither.tools.flattening.__main__:main",
            "slither-format = slither.tools.slither_format.__main__:main",
            "slither-check-erc = slither.tools.erc_conformance.__main__:main",
            "slither-check-kspec = slither.tools.kspec_coverage.__main__:main",
            "slither-prop = slither.tools.properties.__main__:main",
            "slither-mutate = slither.tools.mutator.__main__:main",
            "slither-read-storage = slither.tools.read_storage.__main__:main",
            "slither-doctor = slither.tools.doctor.__main__:main",
            "slither-documentation = slither.tools.documentation.__main__:main",
            "slither-interface = slither.tools.interface.__main__:main",
        ]
    },
)



from __future__ import annotations
from typing import List, Optional, Union
from dataclasses import dataclass


@dataclass
class ASTNode:
    src: str
    node_id: int


@dataclass
class Definition(ASTNode):
    doc_string: Optional[str]


@dataclass
class Module(Definition):
    body: List[ASTNode]
    name: str


@dataclass
class ImportFrom(ASTNode):
    module: str
    name: str
    alias: Optional[str]


@dataclass
class EventDef(ASTNode):
    name: str
    body: List[AnnAssign]


@dataclass
class AnnAssign(ASTNode):
    target: Name
    annotation: Union[Subscript, Name, Call]
    value: Optional[ASTNode]


@dataclass
class Name(ASTNode):  # type or identifier
    id: str


@dataclass
class Call(ASTNode):
    func: ASTNode
    args: List[ASTNode]
    keyword: Optional[ASTNode]
    keywords: List[ASTNode]


@dataclass
class Pass(ASTNode):
    pass


@dataclass
class StructDef(ASTNode):
    name: str
    body: List[AnnAssign]


@dataclass
class VariableDecl(ASTNode):
    annotation: ASTNode
    target: ASTNode
    value: Optional[ASTNode]
    is_constant: bool
    is_immutable: bool
    is_public: bool


@dataclass
class Subscript(ASTNode):
    value: ASTNode
    slice: ASTNode


@dataclass
class Index(ASTNode):
    value: ASTNode


@dataclass
class Bytes(ASTNode):
    value: bytes


@dataclass
class Hex(ASTNode):
    value: str


@dataclass
class Int(ASTNode):
    value: int


@dataclass
class Str(ASTNode):
    value: str


@dataclass
class VyList(ASTNode):
    elements: List[ASTNode]


@dataclass
class VyDict(ASTNode):
    keys: List[ASTNode]
    values: List[ASTNode]


@dataclass
class Tuple(ASTNode):
    elements: List[ASTNode]


@dataclass
class FunctionDef(Definition):
    name: str
    args: Optional[Arguments]
    returns: Optional[List[ASTNode]]
    body: List[ASTNode]
    decorators: Optional[List[ASTNode]]
    pos: Optional[any]  # not sure what this is


@dataclass
class Assign(ASTNode):
    target: ASTNode
    value: ASTNode


@dataclass
class Attribute(ASTNode):
    value: ASTNode
    attr: str


@dataclass
class Arguments(ASTNode):
    args: List[Arg]
    default: Optional[ASTNode]
    defaults: List[ASTNode]


@dataclass
class Arg(ASTNode):
    arg: str
    annotation: Optional[ASTNode]


@dataclass
class Assert(ASTNode):
    test: ASTNode
    msg: Optional[Str]


@dataclass
class Raise(ASTNode):
    exc: ASTNode


@dataclass
class Expr(ASTNode):
    value: ASTNode


@dataclass
class UnaryOp(ASTNode):
    op: ASTNode
    operand: ASTNode


@dataclass
class BinOp(ASTNode):
    left: ASTNode
    op: str
    right: ASTNode


@dataclass
class Keyword(ASTNode):
    arg: str
    value: ASTNode


@dataclass
class Log(ASTNode):
    value: ASTNode


@dataclass
class Return(ASTNode):
    value: Optional[ASTNode]


@dataclass
class If(ASTNode):
    test: ASTNode
    body: List[ASTNode]
    orelse: List[ASTNode]


@dataclass
class Compare(ASTNode):
    left: ASTNode
    op: ASTNode
    right: ASTNode


@dataclass
class NameConstant(ASTNode):
    value: bool


@dataclass
class For(ASTNode):
    target: ASTNode
    iter: ASTNode
    body: List[ASTNode]


@dataclass
class Continue(ASTNode):
    pass


@dataclass
class Break(ASTNode):
    pass


@dataclass
class InterfaceDef(ASTNode):
    name: str
    body: List[ASTNode]


@dataclass
class EnumDef(ASTNode):
    name: str
    body: List[ASTNode]


@dataclass
class AugAssign(ASTNode):
    target: ASTNode
    op: ASTNode
    value: ASTNode


@dataclass
class BoolOp(ASTNode):
    op: ASTNode
    values: List[ASTNode]



from typing import List
from crytic_compile import CryticCompile
from slither.core.declarations import Contract, Function, Enum, Event, Import, Pragma, Structure
from slither.core.solidity_types.type import Type
from slither.core.source_mapping.source_mapping import Source, SourceMapping
from slither.core.variables.variable import Variable
from slither.exceptions import SlitherError


def get_definition(target: SourceMapping, crytic_compile: CryticCompile) -> Source:
    if isinstance(target, (Contract, Function, Enum, Event, Structure, Variable)):
        # Add " " to look after the first solidity keyword
        pattern = " " + target.name
    elif isinstance(target, Import):
        pattern = "import"
    elif isinstance(target, Pragma):
        pattern = "pragma"  # todo maybe return with the while pragma statement
    elif isinstance(target, Type):
        raise SlitherError("get_definition_generic not implemented for types")
    else:
        raise SlitherError(f"get_definition_generic not implemented for {type(target)}")

    file_content = crytic_compile.src_content_for_file(target.source_mapping.filename.absolute)
    txt = file_content[
        target.source_mapping.start : target.source_mapping.start + target.source_mapping.length
    ]

    start_offset = txt.find(pattern) + 1  # remove the space

    starting_line, starting_column = crytic_compile.get_line_from_offset(
        target.source_mapping.filename, target.source_mapping.start + start_offset
    )

    ending_line, ending_column = crytic_compile.get_line_from_offset(
        target.source_mapping.filename, target.source_mapping.start + start_offset + len(pattern)
    )

    s = Source(target.source_mapping.compilation_unit)
    s.start = target.source_mapping.start + start_offset
    s.length = len(pattern)
    s.filename = target.source_mapping.filename
    s.is_dependency = target.source_mapping.is_dependency
    s.lines = list(range(starting_line, ending_line + 1))
    s.starting_column = starting_column
    s.ending_column = ending_column
    s.end = s.start + s.length
    s.txt = txt
    return s


def get_implementation(target: SourceMapping) -> Source:
    return target.source_mapping


def get_references(target: SourceMapping) -> List[Source]:
    return target.references



""""
    Re-entrancy detection

    Based on heuristics, it may lead to FP and FN
    Iterate over all the nodes of the graph until reaching a fixpoint
"""
from collections import namedtuple, defaultdict
from typing import DefaultDict, List, Union, Set

from slither.core.variables.variable import Variable
from slither.detectors.abstract_detector import DetectorClassification
from slither.detectors.reentrancy.reentrancy import Reentrancy, to_hashable
from slither.slithir.operations import Send, Transfer, EventCall
from slither.slithir.operations.high_level_call import HighLevelCall
from slither.slithir.operations.member import Member
from slither.slithir.operations.return_operation import Return
from slither.utils.output import Output

FindingKey = namedtuple("FindingKey", ["function", "calls", "send_eth"])
FindingValue = namedtuple("FindingValue", ["variable", "node", "nodes"])


class ReentrancyNoGas(Reentrancy):
    KEY = "REENTRANCY_NO_GAS"

    ARGUMENT = "reentrancy-unlimited-gas"
    HELP = "Reentrancy vulnerabilities through send and transfer"
    IMPACT = DetectorClassification.INFORMATIONAL
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = (
        "https://github.com/crytic/slither/wiki/Detector-Documentation#reentrancy-vulnerabilities-4"
    )

    WIKI_TITLE = "Reentrancy vulnerabilities"

    # region wiki_description
    WIKI_DESCRIPTION = """
Detection of the [reentrancy bug](https://github.com/trailofbits/not-so-smart-contracts/tree/master/reentrancy).
Only report reentrancy that is based on `transfer` or `send`."""
    # endregion wiki_description

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
    function callme(){
        msg.sender.transfer(balances[msg.sender]):
        balances[msg.sender] = 0;
    }   
```

`send` and `transfer` do not protect from reentrancies in case of gas price changes."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Apply the [`check-effects-interactions` pattern](http://solidity.readthedocs.io/en/v0.4.21/security-considerations.html#re-entrancy)."

    @staticmethod
    def can_callback(ir: Union[Member, Return, HighLevelCall]) -> bool:
        """
        Same as Reentrancy, but also consider Send and Transfer

        """
        return isinstance(ir, (Send, Transfer))

    STANDARD_JSON = False

    def find_reentrancies(self) -> DefaultDict[FindingKey, Set[FindingValue]]:
        result: DefaultDict[FindingKey, Set[FindingValue]] = defaultdict(set)
        for contract in self.contracts:
            for f in contract.functions_and_modifiers_declared:
                for node in f.nodes:
                    # dead code
                    if self.KEY not in node.context:
                        continue
                    if node.context[self.KEY].calls:
                        if not any(n != node for n in node.context[self.KEY].calls):
                            continue

                        # calls are ordered
                        finding_key = FindingKey(
                            function=node.function,
                            calls=to_hashable(node.context[self.KEY].calls),
                            send_eth=to_hashable(node.context[self.KEY].send_eth),
                        )
                        finding_vars = {
                            FindingValue(
                                v,
                                node,
                                tuple(sorted(nodes, key=lambda x: x.node_id)),
                            )
                            for (v, nodes) in node.context[self.KEY].written.items()
                        }
                        finding_vars |= {
                            FindingValue(
                                e,
                                e.node,
                                tuple(sorted(nodes, key=lambda x: x.node_id)),
                            )
                            for (e, nodes) in node.context[self.KEY].events.items()
                        }
                        if finding_vars:
                            result[finding_key] |= finding_vars
        return result

    def _detect(self) -> List[Output]:  # pylint: disable=too-many-branches,too-many-locals
        """"""

        super()._detect()
        reentrancies = self.find_reentrancies()

        results = []

        result_sorted = sorted(list(reentrancies.items()), key=lambda x: x[0][0].name)
        for (func, calls, send_eth), varsWrittenOrEvent in result_sorted:
            calls = sorted(list(set(calls)), key=lambda x: x[0].node_id)
            send_eth = sorted(list(set(send_eth)), key=lambda x: x[0].node_id)
            info = ["Reentrancy in ", func, ":\n"]

            info += ["\tExternal calls:\n"]
            for (call_info, calls_list) in calls:
                info += ["\t- ", call_info, "\n"]
                for call_list_info in calls_list:
                    if call_list_info != call_info:
                        info += ["\t\t- ", call_list_info, "\n"]
            if calls != send_eth and send_eth:
                info += ["\tExternal calls sending eth:\n"]
                for (call_info, calls_list) in send_eth:
                    info += ["\t- ", call_info, "\n"]
                    for call_list_info in calls_list:
                        if call_list_info != call_info:
                            info += ["\t\t- ", call_list_info, "\n"]

            varsWritten = [
                FindingValue(v, node, nodes)
                for (v, node, nodes) in varsWrittenOrEvent
                if isinstance(v, Variable)
            ]
            varsWritten = sorted(varsWritten, key=lambda x: (x.variable.name, x.node.node_id))
            if varsWritten:
                info += ["\tState variables written after the call(s):\n"]
                for finding_value in varsWritten:
                    info += ["\t- ", finding_value.node, "\n"]
                    for other_node in finding_value.nodes:
                        if other_node != finding_value.node:
                            info += ["\t\t- ", other_node, "\n"]

            events = [
                FindingValue(v, node, nodes)
                for (v, node, nodes) in varsWrittenOrEvent
                if isinstance(v, EventCall)
            ]
            events = sorted(events, key=lambda x: (x.variable.name, x.node.node_id))
            if events:
                info += ["\tEvent emitted after the call(s):\n"]
                for finding_value in events:
                    info += ["\t- ", finding_value.node, "\n"]
                    for other_node in finding_value.nodes:
                        if other_node != finding_value.node:
                            info += ["\t\t- ", other_node, "\n"]

            # Create our JSON result
            res = self.generate_result(info)

            # Add the function with the re-entrancy first
            res.add(func)

            # Add all underlying calls in the function which are potentially problematic.
            for (call_info, calls_list) in calls:
                res.add(call_info, {"underlying_type": "external_calls"})
                for call_list_info in calls_list:
                    if call_list_info != call_info:
                        res.add(
                            call_list_info,
                            {"underlying_type": "external_calls_sending_eth"},
                        )

            #

            # If the calls are not the same ones that send eth, add the eth sending nodes.
            if calls != send_eth:
                for (call_info, calls_list) in send_eth:
                    res.add(call_info, {"underlying_type": "external_calls_sending_eth"})
                    for call_list_info in calls_list:
                        if call_list_info != call_info:
                            res.add(
                                call_list_info,
                                {"underlying_type": "external_calls_sending_eth"},
                            )

            # Add all variables written via nodes which write them.
            for finding_value in varsWritten:
                res.add(
                    finding_value.node,
                    {
                        "underlying_type": "variables_written",
                        "variable_name": finding_value.variable.name,
                    },
                )
                for other_node in finding_value.nodes:
                    if other_node != finding_value.node:
                        res.add(
                            other_node,
                            {
                                "underlying_type": "variables_written",
                                "variable_name": finding_value.variable.name,
                            },
                        )
            for finding_value in events:
                res.add(finding_value.node, {"underlying_type": "event"})
                for other_node in finding_value.nodes:
                    if other_node != finding_value.node:
                        res.add(other_node, {"underlying_type": "event"})

            # Append our result
            results.append(res)

        return results



"""
    Module printing evm mapping of the contract
"""
from slither.printers.abstract_printer import AbstractPrinter
from slither.analyses.evm import (
    generate_source_to_evm_ins_mapping,
    load_evm_cfg_builder,
)
from slither.utils.colors import blue, green, magenta, red


def _extract_evm_info(slither):
    """
    Extract evm information for all derived contracts using evm_cfg_builder

    Returns: evm CFG and Solidity source to Program Counter (pc) mapping
    """

    evm_info = {}

    CFG = load_evm_cfg_builder()

    for contract in slither.contracts_derived:
        contract_bytecode_runtime = contract.file_scope.bytecode_runtime(
            contract.compilation_unit.crytic_compile_compilation_unit, contract.name
        )
        contract_srcmap_runtime = contract.file_scope.srcmap_runtime(
            contract.compilation_unit.crytic_compile_compilation_unit, contract.name
        )
        cfg = CFG(contract_bytecode_runtime)
        evm_info["cfg", contract.name] = cfg
        evm_info["mapping", contract.name] = generate_source_to_evm_ins_mapping(
            cfg.instructions,
            contract_srcmap_runtime,
            slither,
            contract.source_mapping.filename.absolute,
        )

        contract_bytecode_init = contract.file_scope.bytecode_init(
            contract.compilation_unit.crytic_compile_compilation_unit, contract.name
        )
        contract_srcmap_init = contract.file_scope.srcmap_init(
            contract.compilation_unit.crytic_compile_compilation_unit, contract.name
        )
        cfg_init = CFG(contract_bytecode_init)

        evm_info["cfg_init", contract.name] = cfg_init
        evm_info["mapping_init", contract.name] = generate_source_to_evm_ins_mapping(
            cfg_init.instructions,
            contract_srcmap_init,
            slither,
            contract.source_mapping.filename.absolute,
        )

    return evm_info


# pylint: disable=too-many-locals
class PrinterEVM(AbstractPrinter):
    ARGUMENT = "evm"
    HELP = "Print the evm instructions of nodes in functions"

    WIKI = "https://github.com/trailofbits/slither/wiki/Printer-documentation#evm"

    def output(self, _filename):
        """
        _filename is not used
        Args:
            _filename(string)
        """

        txt = ""
        if not self.slither.crytic_compile:
            txt = "The EVM printer requires to compile with crytic-compile"
            self.info(red(txt))
            res = self.generate_output(txt)
            return res
        evm_info = _extract_evm_info(self.slither)

        for contract in self.slither.contracts_derived:
            txt += blue(f"Contract {contract.name}\n")

            contract_file = self.slither.source_code[
                contract.source_mapping.filename.absolute
            ].encode("utf-8")
            with open(contract.source_mapping.filename.absolute, "r", encoding="utf8") as f:
                contract_file_lines = f.readlines()

            contract_pcs = {}
            contract_cfg = {}

            for function in contract.functions:
                txt += blue(f"\tFunction {function.canonical_name}\n")

                # CFG and source mapping depend on function being constructor or not
                if function.is_constructor:
                    contract_cfg = evm_info["cfg_init", contract.name]
                    contract_pcs = evm_info["mapping_init", contract.name]
                else:
                    contract_cfg = evm_info["cfg", contract.name]
                    contract_pcs = evm_info["mapping", contract.name]

                for node in function.nodes:
                    txt += green("\t\tNode: " + str(node) + "\n")
                    node_source_line = (
                        contract_file[0 : node.source_mapping.start].count("\n".encode("utf-8")) + 1
                    )
                    txt += green(
                        f"\t\tSource line {node_source_line}: {contract_file_lines[node_source_line - 1].rstrip()}\n"
                    )
                    txt += magenta("\t\tEVM Instructions:\n")
                    node_pcs = contract_pcs.get(node_source_line, [])
                    for pc in node_pcs:
                        txt += magenta(f"\t\t\t{hex(pc)}: {contract_cfg.get_instruction_at(pc)}\n")

            for modifier in contract.modifiers:
                txt += blue(f"\tModifier {modifier.canonical_name}\n")
                for node in modifier.nodes:
                    txt += green("\t\tNode: " + str(node) + "\n")
                    node_source_line = (
                        contract_file[0 : node.source_mapping.start].count("\n".encode("utf-8")) + 1
                    )
                    txt += green(
                        f"\t\tSource line {node_source_line}: {contract_file_lines[node_source_line - 1].rstrip()}\n"
                    )
                    txt += magenta("\t\tEVM Instructions:\n")
                    node_pcs = contract_pcs.get(node_source_line, [])
                    for pc in node_pcs:
                        txt += magenta(f"\t\t\t{hex(pc)}: {contract_cfg.get_instruction_at(pc)}\n")

        self.info(txt)
        res = self.generate_output(txt)
        return res



"""
    Structure module
"""
from typing import TYPE_CHECKING, Dict

from slither.core.declarations.structure import Structure
from slither.core.variables.structure_variable import StructureVariable
from slither.solc_parsing.variables.structure_variable import StructureVariableSolc

if TYPE_CHECKING:
    from slither.solc_parsing.declarations.contract import ContractSolc


class StructureContractSolc:  # pylint: disable=too-few-public-methods
    """
    Structure class
    """

    # elems = [(type, name)]

    def __init__(  # pylint: disable=too-many-arguments
        self,
        st: Structure,
        struct: Dict,
        contract_parser: "ContractSolc",
    ) -> None:

        if contract_parser.is_compact_ast:
            name = struct["name"]
            attributes = struct
        else:
            name = struct["attributes"][contract_parser.get_key()]
            attributes = struct["attributes"]
        if "canonicalName" in attributes:
            canonicalName = attributes["canonicalName"]
        else:
            canonicalName = contract_parser.underlying_contract.name + "." + name

        children = struct["members"] if "members" in struct else struct.get("children", [])

        self._structure = st
        st.name = name
        st.canonical_name = canonicalName
        self._contract_parser = contract_parser

        self._elemsNotParsed = children

    def analyze(self) -> None:
        for elem_to_parse in self._elemsNotParsed:
            elem = StructureVariable()
            elem.set_structure(self._structure)
            elem.set_offset(elem_to_parse["src"], self._structure.contract.compilation_unit)

            elem_parser = StructureVariableSolc(elem, elem_to_parse)
            elem_parser.analyze(self._contract_parser)

            self._structure.elems[elem.name] = elem
            self._structure.add_elem_in_order(elem.name)
        self._elemsNotParsed = []



import logging
from typing import Union, List, ValuesView, Type, Dict, Optional

from crytic_compile import CryticCompile, InvalidCompilation

# pylint: disable= no-name-in-module
from slither.core.compilation_unit import SlitherCompilationUnit
from slither.core.scope.scope import FileScope
from slither.core.slither_core import SlitherCore
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.exceptions import SlitherError
from slither.printers.abstract_printer import AbstractPrinter
from slither.solc_parsing.slither_compilation_unit_solc import SlitherCompilationUnitSolc
from slither.vyper_parsing.vyper_compilation_unit import VyperCompilationUnit
from slither.utils.output import Output
from slither.vyper_parsing.ast.ast import parse

logger = logging.getLogger("Slither")
logging.basicConfig()

logger_detector = logging.getLogger("Detectors")
logger_printer = logging.getLogger("Printers")


def _check_common_things(
    thing_name: str, cls: Type, base_cls: Type, instances_list: List[Type[AbstractDetector]]
) -> None:

    if not issubclass(cls, base_cls) or cls is base_cls:
        raise Exception(
            f"You can't register {cls!r} as a {thing_name}. You need to pass a class that inherits from {base_cls.__name__}"
        )

    if any(type(obj) == cls for obj in instances_list):  # pylint: disable=unidiomatic-typecheck
        raise Exception(f"You can't register {cls!r} twice.")


def _update_file_scopes(candidates: ValuesView[FileScope]):
    """
    Because solc's import allows cycle in the import
    We iterate until we aren't adding new information to the scope

    """
    learned_something = False
    while True:
        for candidate in candidates:
            learned_something |= candidate.add_accesible_scopes()
        if not learned_something:
            break
        learned_something = False


class Slither(
    SlitherCore
):  # pylint: disable=too-many-instance-attributes,too-many-locals,too-many-statements
    def __init__(self, target: Union[str, CryticCompile], **kwargs) -> None:
        """
        Args:
            target (str | CryticCompile)
        Keyword Args:
            solc (str): solc binary location (default 'solc')
            disable_solc_warnings (bool): True to disable solc warnings (default false)
            solc_args (str): solc arguments (default '')
            ast_format (str): ast format (default '--ast-compact-json')
            filter_paths (list(str)): list of path to filter (default [])
            triage_mode (bool): if true, switch to triage mode (default false)
            exclude_dependencies (bool): if true, exclude results that are only related to dependencies
            generate_patches (bool): if true, patches are generated (json output only)
            change_line_prefix (str): Change the line prefix (default #)
                for the displayed source codes (i.e. file.sol#1).

        """
        super().__init__()

        self._disallow_partial: bool = kwargs.get("disallow_partial", False)
        self._skip_assembly: bool = kwargs.get("skip_assembly", False)
        self._show_ignored_findings: bool = kwargs.get("show_ignored_findings", False)

        self.line_prefix = kwargs.get("change_line_prefix", "#")

        # Indicate if Codex related features should be used
        self.codex_enabled = kwargs.get("codex", False)
        self.codex_contracts = kwargs.get("codex_contracts", "all")
        self.codex_model = kwargs.get("codex_model", "text-davinci-003")
        self.codex_temperature = kwargs.get("codex_temperature", 0)
        self.codex_max_tokens = kwargs.get("codex_max_tokens", 300)
        self.codex_log = kwargs.get("codex_log", False)
        self.codex_organization: Optional[str] = kwargs.get("codex_organization", None)

        self.no_fail = kwargs.get("no_fail", False)

        self._parsers: List[SlitherCompilationUnitSolc] = []
        try:
            if isinstance(target, CryticCompile):
                crytic_compile = target
            else:
                crytic_compile = CryticCompile(target, **kwargs)
            self._crytic_compile = crytic_compile
        except InvalidCompilation as e:
            # pylint: disable=raise-missing-from
            raise SlitherError(f"Invalid compilation: \n{str(e)}")
        for compilation_unit in crytic_compile.compilation_units.values():
            compilation_unit_slither = SlitherCompilationUnit(self, compilation_unit)
            self._compilation_units.append(compilation_unit_slither)

            if compilation_unit_slither.is_vyper:
                vyper_parser = VyperCompilationUnit(compilation_unit_slither)
                for path, ast in compilation_unit.asts.items():
                    ast_nodes = parse(ast["ast"])
                    vyper_parser.parse_module(ast_nodes, path)
                self._parsers.append(vyper_parser)
            else:
                # Solidity specific
                assert compilation_unit_slither.is_solidity
                sol_parser = SlitherCompilationUnitSolc(compilation_unit_slither)
                self._parsers.append(sol_parser)
                for path, ast in compilation_unit.asts.items():
                    sol_parser.parse_top_level_items(ast, path)
                    self.add_source_code(path)

                _update_file_scopes(compilation_unit_slither.scopes.values())

        if kwargs.get("generate_patches", False):
            self.generate_patches = True

        self._markdown_root = kwargs.get("markdown_root", "")

        self._detectors = []
        self._printers = []

        filter_paths = kwargs.get("filter_paths", [])
        for p in filter_paths:
            self.add_path_to_filter(p)

        include_paths = kwargs.get("include_paths", [])
        for p in include_paths:
            self.add_path_to_include(p)

        self._exclude_dependencies = kwargs.get("exclude_dependencies", False)

        triage_mode = kwargs.get("triage_mode", False)
        triage_database = kwargs.get("triage_database", "slither.db.json")
        self._triage_mode = triage_mode
        self._previous_results_filename = triage_database

        printers_to_run = kwargs.get("printers_to_run", "")
        if printers_to_run == "echidna":
            self.skip_data_dependency = True

        self._init_parsing_and_analyses(kwargs.get("skip_analyze", False))

    def _init_parsing_and_analyses(self, skip_analyze: bool) -> None:

        for parser in self._parsers:
            try:
                parser.parse_contracts()
            except Exception as e:
                if self.no_fail:
                    continue
                raise e

        # skip_analyze is only used for testing
        if not skip_analyze:
            for parser in self._parsers:
                try:
                    parser.analyze_contracts()
                except Exception as e:
                    if self.no_fail:
                        continue
                    raise e

    @property
    def detectors(self):
        return self._detectors

    @property
    def detectors_high(self):
        return [d for d in self.detectors if d.IMPACT == DetectorClassification.HIGH]

    @property
    def detectors_medium(self):
        return [d for d in self.detectors if d.IMPACT == DetectorClassification.MEDIUM]

    @property
    def detectors_low(self):
        return [d for d in self.detectors if d.IMPACT == DetectorClassification.LOW]

    @property
    def detectors_informational(self):
        return [d for d in self.detectors if d.IMPACT == DetectorClassification.INFORMATIONAL]

    @property
    def detectors_optimization(self):
        return [d for d in self.detectors if d.IMPACT == DetectorClassification.OPTIMIZATION]

    def register_detector(self, detector_class: Type[AbstractDetector]) -> None:
        """
        :param detector_class: Class inheriting from `AbstractDetector`.
        """
        _check_common_things("detector", detector_class, AbstractDetector, self._detectors)

        for compilation_unit in self.compilation_units:
            instance = detector_class(compilation_unit, self, logger_detector)
            self._detectors.append(instance)

    def unregister_detector(self, detector_class: Type[AbstractDetector]) -> None:
        """
        :param detector_class: Class inheriting from `AbstractDetector`.
        """

        for obj in self._detectors:
            if isinstance(obj, detector_class):
                self._detectors.remove(obj)
                return

    def register_printer(self, printer_class: Type[AbstractPrinter]) -> None:
        """
        :param printer_class: Class inheriting from `AbstractPrinter`.
        """
        _check_common_things("printer", printer_class, AbstractPrinter, self._printers)

        instance = printer_class(self, logger_printer)
        self._printers.append(instance)

    def unregister_printer(self, printer_class: Type[AbstractPrinter]) -> None:
        """
        :param printer_class: Class inheriting from `AbstractPrinter`.
        """

        for obj in self._printers:
            if isinstance(obj, printer_class):
                self._printers.remove(obj)
                return

    def run_detectors(self) -> List[Dict]:
        """
        :return: List of registered detectors results.
        """

        self.load_previous_results()
        results = [d.detect() for d in self._detectors]

        self.write_results_to_hide()
        return results

    def run_printers(self) -> List[Output]:
        """
        :return: List of registered printers outputs.
        """

        return [p.output(self._crytic_compile.target).data for p in self._printers]

    @property
    def triage_mode(self) -> bool:
        return self._triage_mode
